! Copyright (C) 2002, 2007 Carnegie Mellon University and others.
! All Rights Reserved.
! This code is published under the Common Public License.
!
! =============================================================================
!
!  This is an example for the usage of IPOPT.
!  It implements problem 71 from the Hock-Schittkowski test suite:
!
!  min   x1*x4*(x1 + x2 + x3)  +  x3
!  s.t.  x1*x2*x3*x4                   >=  25
!        x1**2 + x2**2 + x3**2 + x4**2  =  40
!        1 <=  x1,x2,x3,x4  <= 5
!
!  Starting point:
!     x = (1, 5, 5, 1)
!
!  Optimal solution:
!     x = (1.00000000, 4.74299963, 3.82114998, 1.37940829)
!
! =============================================================================
!
!
! =============================================================================
!
!                         Main driver program
!
! =============================================================================
!
program example
  !
  implicit none
  !
  !  include the Ipopt return codes
  !
  include 'IpReturnCodes.inc'
  !
  !  Size of the problem (number of variables and equality constraints)
  !
  integer     N,     M,     NELE_JAC,     NELE_HESS,      IDX_STY
  parameter  (N = 4, M = 2, NELE_JAC = 8, NELE_HESS = 10)
  parameter  (IDX_STY = 1 )
  !
  !  Space for multipliers and constraints
  !
  double precision LAM(M)
  double precision G(M)
  !
  !  Vector of variables
  !
  double precision X(N)
  !
  !  Vector of lower and upper bounds
  !
  double precision X_L(N), X_U(N), Z_L(N), Z_U(N)
  double precision G_L(M), G_U(M)
  !
  !  Private data for evaluation routines
  !  This could be used to pass double precision and integer arrays untouched
  !  to the evaluation subroutines EVAL_*
  !
  double precision DAT(2)
  integer IDAT(1)
  !
  !  Place for storing the Ipopt Problem Handle
  !
  !  for 32 bit platforms
  integer IPROBLEM
  integer IPCREATE
  !
  integer IERR
  integer IPSOLVE, IPADDSTROPTION
  integer IPADDNUMOPTION, IPADDINTOPTION
  integer IPOPENOUTPUTFILE
  !
  double precision F
  integer i
  !
  !  The following are the Fortran routines for computing the model
  !  functions and their derivatives - their code can be found furhter
  !  down in this file.
  !
  external EV_F, EV_G, EV_GRAD_F, EV_JAC_G, EV_HESS
  !
  !  Set initial point and bounds:
  !
  data X   / 1d0, 5d0, 5d0, 1d0/
  data X_L / 1d0, 1d0, 1d0, 1d0 /
  data X_U / 5d0, 5d0, 5d0, 5d0 /
  !
  !  Set bounds for the constraints
  !
  data G_L / 25d0, 40d0 /
  data G_U / 1d40, 40d0 /
  !
  !  First create a handle for the Ipopt problem (and read the options
  !  file)
  !
  IPROBLEM = IPCREATE(N, X_L, X_U, M, G_L, G_U, NELE_JAC, NELE_HESS,IDX_STY, EV_F, EV_G, EV_GRAD_F, EV_JAC_G, EV_HESS)
  if (IPROBLEM.eq.0) then
     write(*,*) 'Error creating an Ipopt Problem handle.'
     stop
  endif
  !
  !  Open an output file
  !
  IERR = IPOPENOUTPUTFILE(IPROBLEM, 'IPOPT.OUT', 5)
  if (IERR.ne.0 ) then
     write(*,*) 'Error opening the Ipopt output file.'
     goto 9000
  endif
  !
  !  Note: The following options are only examples, they might not be
  !        suitable for your optimization problem.
  !
  !  Set a string option
  !
  IERR = IPADDSTROPTION(IPROBLEM, 'mu_strategy', 'adaptive')
  if (IERR.ne.0 ) goto 9990
  !
  !  Set an integer option
  !
  IERR = IPADDINTOPTION(IPROBLEM, 'max_iter', 3000)
  if (IERR.ne.0 ) goto 9990
  !
  !  Set a double precision option
  !
  IERR = IPADDNUMOPTION(IPROBLEM, 'tol', 1.d-7)
  if (IERR.ne.0 ) goto 9990
  !
  !  As a simple example, we pass the constants in the constraints to
  !  the EVAL_C routine via the "private" DAT array.
  !
  DAT(1) = 0.d0
  DAT(2) = 0.d0
  !
  !  Call optimization routine
  !
  IERR = IPSOLVE(IPROBLEM, X, G, F, LAM, Z_L, Z_U, IDAT, DAT)
  !
  !  Output:
  !
  if( IERR.eq.IP_SOLVE_SUCCEEDED ) then
     write(*,*)
     write(*,*) 'The solution was found.'
     write(*,*)
     write(*,*) 'The final value of the objective function is ',F
     write(*,*)
     write(*,*) 'The optimal values of X are:'
     write(*,*)
     do i = 1, N
        write(*,*) 'X  (',i,') = ',X(i)
     enddo
     write(*,*)
     write(*,*) 'The multipliers for the lower bounds are:'
     write(*,*)
     do i = 1, N
        write(*,*) 'Z_L(',i,') = ',Z_L(i)
     enddo
     write(*,*)
     write(*,*) 'The multipliers for the upper bounds are:'
     write(*,*)
     do i = 1, N
        write(*,*) 'Z_U(',i,') = ',Z_U(i)
     enddo
     write(*,*)
     write(*,*) 'The multipliers for the equality constraints are:'
     write(*,*)
     do i = 1, M
        write(*,*) 'LAM(',i,') = ',LAM(i)
     enddo
     write(*,*)
  else
     write(*,*)
     write(*,*) 'An error occoured.'
     write(*,*) 'The error code is ',IERR
     write(*,*)
  endif
  !
9000 continue
  !
  !  Clean up
  !
  call IPFREE(IPROBLEM)
  stop
  !
9990 continue
  write(*,*) 'Error setting an option'
  goto 9000
end program example
!
! =============================================================================
!
!                 Computation of objective function
!
! =============================================================================
!
subroutine EV_F(N, X, NEW_X, F, IDAT, DAT, IERR)
  implicit none
  integer N, NEW_X
  double precision F, X(N)
  double precision DAT(*)
  integer IDAT(*)
  integer IERR
  F = X(1)*X(4)*(X(1)+X(2)+X(3)) + X(3)
  IERR = 0
  return
end subroutine EV_F
!
! =============================================================================
!
!             Computation of gradient of objective function
!
! =============================================================================
!
subroutine EV_GRAD_F(N, X, NEW_X, GRAD, IDAT, DAT, IERR)
  implicit none
  integer N, NEW_X
  double precision GRAD(N), X(N)
  double precision DAT(*)
  integer IDAT(*)
  integer IERR
  GRAD(1) = X(4)*(2d0*X(1)+X(2)+X(3))
  GRAD(2) = X(1)*X(4)
  GRAD(3) = X(1)*X(4) + 1d0
  GRAD(4) = X(1)*(X(1)+X(2)+X(3))
  IERR = 0
  return
end subroutine EV_GRAD_F
!
! =============================================================================
!
!                  Computation of equality constraints
!
! =============================================================================
!
subroutine EV_G(N, X, NEW_X, M, G, IDAT, DAT, IERR)
  implicit none
  integer N, NEW_X, M
  double precision G(M), X(N)
  double precision DAT(*)
  integer IDAT(*)
  integer IERR
  G(1) = X(1)*X(2)*X(3)*X(4) - DAT(1)
  G(2) = X(1)**2 + X(2)**2 + X(3)**2 + X(4)**2 - DAT(2)
  IERR = 0
  return
end subroutine EV_G
!
! =============================================================================
!
!             Computation of Jacobian of equality constraints
!
! =============================================================================
!
subroutine EV_JAC_G(TASK, N, X, NEW_X, M, NZ, ACON, AVAR, A, IDAT, DAT, IERR)
  integer TASK, N, NEW_X, M, NZ
  double precision X(N), A(NZ)
  integer ACON(NZ), AVAR(NZ), I
  double precision DAT(*)
  integer IDAT(*)
  integer IERR
  !
  !  structure of Jacobian:
  !
  integer AVAR1(8), ACON1(8)
  data  AVAR1 /1, 2, 3, 4, 1, 2, 3, 4/
  data  ACON1 /1, 1, 1, 1, 2, 2, 2, 2/
  save  AVAR1, ACON1
  !
  if( TASK.eq.0 ) then
     do I = 1, 8
        AVAR(I) = AVAR1(I)
        ACON(I) = ACON1(I)
     enddo
  else
     A(1) = X(2)*X(3)*X(4)
     A(2) = X(1)*X(3)*X(4)
     A(3) = X(1)*X(2)*X(4)
     A(4) = X(1)*X(2)*X(3)
     A(5) = 2d0*X(1)
     A(6) = 2d0*X(2)
     A(7) = 2d0*X(3)
     A(8) = 2d0*X(4)
  endif
  IERR = 0
  return
end subroutine EV_JAC_G
!
! =============================================================================
!
!             Computation of Hessian of Lagrangian
!
! =============================================================================
!
subroutine EV_HESS(TASK, N, X, NEW_X, OBJFACT, M, LAM, NEW_LAM, NNZH, IRNH, ICNH, HESS, IDAT, DAT, IERR)
  implicit none
  integer TASK, N, NEW_X, M, NEW_LAM, NNZH, i
  double precision X(N), OBJFACT, LAM(M), HESS(NNZH)
  integer IRNH(NNZH), ICNH(NNZH)
  double precision DAT(*)
  integer IDAT(*)
  integer IERR
  !
  !  structure of Hessian:
  !
  integer IRNH1(10), ICNH1(10)
  data  IRNH1 /1, 2, 2, 3, 3, 3, 4, 4, 4, 4/
  data  ICNH1 /1, 1, 2, 1, 2, 3, 1, 2, 3, 4/
  save  IRNH1, ICNH1

  if( TASK.eq.0 ) then
     do i = 1, 10
        IRNH(i) = IRNH1(i)
        ICNH(i) = ICNH1(i)
     enddo
  else
     do i = 1, 10
        HESS(i) = 0d0
     enddo
     !
     !  objective function
     !
     HESS(1) = OBJFACT * 2d0*X(4)
     HESS(2) = OBJFACT * X(4)
     HESS(4) = OBJFACT * X(4)
     HESS(7) = OBJFACT * (2d0*X(1) + X(2) + X(3))
     HESS(8) = OBJFACT * X(1)
     HESS(9) = OBJFACT * X(1)
     !
     !  first constraint
     !
     HESS(2) = HESS(2) + LAM(1) * X(3)*X(4)
     HESS(4) = HESS(4) + LAM(1) * X(2)*X(4)
     HESS(5) = HESS(5) + LAM(1) * X(1)*X(4)
     HESS(7) = HESS(7) + LAM(1) * X(2)*X(3)
     HESS(8) = HESS(8) + LAM(1) * X(1)*X(3)
     HESS(9) = HESS(9) + LAM(1) * X(1)*X(2)
     !
     !  second constraint
     !
     HESS(1) = HESS(1) + LAM(2) * 2d0
     HESS(3) = HESS(3) + LAM(2) * 2d0
     HESS(6) = HESS(6) + LAM(2) * 2d0
     HESS(10)= HESS(10)+ LAM(2) * 2d0
  endif
  IERR = 0
  return
end subroutine EV_HESS
